/*-
 * Copyright (C)2020..2021 @BABOLO http://www.babolo.ru/
 * PKG = babolo-libmake
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2020..2021 @BABOLO http://www.babolo.ru/\n"
#ident "@(#) $Id: over.c,v 1.2 2021/08/22 15:57:57 babolo Exp $\n"

#define Bpars_COMPAT  4
#define MULAR_COMPAT  0
#define MIFE_COMPAT   5
#define BLIN_COMPAT   VMAJOR
#define BLIN_INTERNAL

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <limits.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include "../BLINflag.h"
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "blinread.h"

int
BLINr_rq(BLINr_cfg *cfg, BLIN_trace *twind) {
#   define isJX(EXCODE, FORMAT, ...) \
        do { ifBLIN_QX0("!!" FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
#   define isJW(EXCODE, FORMAT, ...) \
        do { ifBLIN_QW0("!!" FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
#   define blin_internal_flags (cfg->flags & BLIN_MASK)
    u_int64_t         fstime;                  /*    tracfile                   */
    int               ex = EX_OK;              /*                                */
    u_int32_t         p;                       /* ޣ,              */

    qbuf.time1 = qbuf.time2 = fstime = twind->time;
    if  (cfg->flags & (BLINREAD_P_D | BLINREAD_P_T)) {
        ifBLIN_QX4("qbuf.time={%"BLIN_U" %"BLIN_U"}", qbuf.time1, qbuf.time2);
    }
    if  (cfg->flags & BLINREAD_P_I) {
        qbuf.pid = malloc(sizeof(BLIN_list_u4));
        if  (!qbuf.pid) isJW(EX_OSERR, "qbuf init pid malloc");
        qbuf.pid->next = NULL;
        qbuf.pid->num = twind->pid;
        ifBLIN_QX4("qbuf.pid=%u", qbuf.pid->num);
    }
    if  (cfg->flags & BLINREAD_P_R) {
        if  (!(qbuf.prog = malloc(sizeof(BLIN_list_pc)))) isJW(EX_OSERR, "qbuf init prog malloc");
        qbuf.prog->next = NULL;
        qbuf.prog->name = strndup(&twind->string[0], twind->snprog);
        if  (!qbuf.prog->name)isJW(EX_OSERR, "qbuf init prog strndup");
        ifBLIN_QX4("qbuf.prog=%s", qbuf.prog->name);
    }
    if  (cfg->flags & (BLINREAD_P_S | BLINREAD_P_U | BLINREAD_P_N)) {
        BLIN_qbuf_list *func;
        BLIN_list_u4   *line;

        if  (!(qbuf.file = malloc(sizeof(BLIN_qbuf_list)))) isJW(EX_OSERR, "qbuf init file malloc");
        qbuf.file->next = NULL;
        qbuf.file->name = strndup(&twind->string[twind->snprog], twind->snfile);
        if  (!qbuf.file->name) isJW(EX_OSERR, "qbuf init file strndup");
        if  (!(func = malloc(sizeof(BLIN_qbuf_list)))) isJW(EX_OSERR, "qbuf init func malloc");
        qbuf.file->list = func;
        func->next = NULL;
        func->name = strndup(&twind->string[twind->snprog + twind->snfile], twind->snfunc);
        if  (!func->name)    isJW(EX_OSERR, "qbuf init func strndup");
        if  (!(line = malloc(sizeof(BLIN_list_u4)))) isJW(EX_OSERR, "qbuf init line malloc");
        func->list = line;
        line->next = NULL;
        line->num  = twind->line;
        ifBLIN_QX4("qbuf.file={%s %s %u}", qbuf.file->name, func->name, line->num);
    }
    for (; ; twind = BLINr_next_wind(cfg)) {  /*****/
        if  (!twind) isJW(EX_IOERR, "tracefile mife_window %"BLIN_U, cfg->tfile->ofpoint);
        ifBLIN_QX3( "twind={%u %u %u %"BLIN_O"u %u %u %u %u %u %.*s}"
                  , twind->len
                  , twind->level
                  , twind->pid
                  , twind->time
                  , twind->erno
                  , twind->line
                  , twind->snprog
                  , twind->snfile
                  , twind->snfunc
                  , (int) (twind->len - offsetof(BLIN_trace, string))
                  , twind->string
                  )
        ;
        qbuf.level |= (0x01 << twind->level);
        if  (cfg->flags & BLINREAD_P_L) ifBLIN_QX4("qbuf.level=%02x", qbuf.level);
        if  (qbuf.time1 > twind->time) {
            ifBLIN_QX1( "Base time shift %"BLIN_O"u.%09u > %"BLIN_O"u.%09u"
                      , qbuf.time1 / BLINr_E9
                      , (u_int32_t)(qbuf.time1 % BLINr_E9)
                      , twind->time / BLINr_E9
                      , (u_int32_t)(twind->time % BLINr_E9)
                      );
            qbuf.time1 = twind->time;
        } else if(qbuf.time2 < twind->time) {
            qbuf.time2 = twind->time;
        }
        if  (cfg->flags & (BLINREAD_P_D | BLINREAD_P_T)) {
            ifBLIN_QX4("qbuf.time={%"BLIN_U" %"BLIN_U"}", qbuf.time1, qbuf.time2);
        }
        if  (qbuf.pid) {
            for (BLIN_list_u4 *i = qbuf.pid; i->num != twind->pid; i = i->next) {
        ;       if  (!i->next) {
        ;           if  (!(i->next = malloc(sizeof(BLIN_list_u4)))) isJW(1, "qbuf pid malloc");
        ;           i->next->num = twind->pid;
        ;           i->next->next = 0;
        ;           ifBLIN_QX4("qbuf new pid %u", i->next->num);
                    break;
        }   }   }
        if  (qbuf.prog) {
            for ( BLIN_list_pc *i = qbuf.prog
                ; strncmp(i->name, &twind->string[0], twind->snprog)
                ; i = i->next
                ) {
        ;       if  (!i->next) {
        ;           if  (!(i->next = malloc(sizeof(BLIN_list_pc)))) isJW(EX_OSERR, "qbuf prog malloc");
        ;           i->next->name = strndup(&twind->string[0], twind->snprog);
        ;           if  (!i->next->name) isJW(EX_OSERR, "qbuf prog strndup");
        ;           i->next->next = 0;
        ;           ifBLIN_QX4("qbuf new prog %s", i->next->name);
                    break;
        }   }   }
        if  (qbuf.file) {  /*****/
            BLIN_qbuf_list *file;
            BLIN_qbuf_list *func;
            BLIN_list_u4   *line;
            BLIN_qbuf_list *qa;
            BLIN_qbuf_list *qam;
            BLIN_list_u4   *qb;
            BLIN_list_u4   *qbm;

        ;   if  (!(file = malloc(sizeof(BLIN_qbuf_list)))) isJW(EX_OSERR, "qbuf init file malloc");
        ;   file->next = NULL;
        ;   file->name = strndup(&twind->string[twind->snprog], twind->snfile);
        ;   if  (!file->name) isJW(EX_OSERR, "qbuf init file strndup");
        ;   if  (!(func = malloc(sizeof(BLIN_qbuf_list)))) isJW(EX_OSERR, "qbuf init func malloc");
        ;   file->list = func;
        ;   func->next = NULL;
        ;   func->name = strndup(&twind->string[twind->snprog + twind->snfile], twind->snfunc);
        ;   if  (!func->name) isJW(EX_OSERR, "qbuf init func strndup");
        ;   if  (!(line = malloc(sizeof(BLIN_list_u4)))) isJW(EX_OSERR, "qbuf init line malloc");
        ;   func->list = line;
        ;   line->next = NULL;
        ;   line->num  = twind->line;
        ;   for ( qam = qa = qbuf.file
                ; (qa->next) && (strncmp(qa->name, file->name, twind->snfile) < 0)
                ; qa = (qam = qa)->next
                ) {}
        ;   ifBLIN_QX4( "strncmp(%s, %s, %u) == %d"
                      , qa->name
                      , file->name
                      , twind->snfile
                      , strncmp(qa->name, file->name, twind->snfile)
                      )
        ;   ;
        ;   if  (strncmp(qa->name, file->name, twind->snfile)) {
        ;       if  (qa->next) {
        ;           file->next = qam->next;
        ;           qam->next = file;
        ;       } else {
        ;           qa->next = file;
        ;       }
        ;       ifBLIN_QX4("qbuf new file %s", file->name);
        ;       goto fqba;
        ;   }
        ;   for ( qam = qa = qa->list
                ; (qa->next) && (strncmp(qa->name, func->name, twind->snfunc) < 0)
                ; qa = (qam = qa)->next
                ) {}
        ;   if  (strncmp(qa->name, func->name, twind->snfunc)) {
        ;       if  (qa->next) {
        ;           func->next = qam->next;
        ;           qam->next = func;
        ;       } else {
        ;           qa->next = func;
        ;       }
        ;       ifBLIN_QX4("qbuf new func %s", func->name);
        ;       goto fqba;
        ;   }
        ;   for (qbm = qb = qa->list; (qb->next) && (qb->num < line->num); qb = (qbm = qb)->next) {}
        ;   if  (qb->num != line->num) {
        ;       if  (qb->next) {
        ;           line->next = qbm->next;
        ;           qbm->next = line;
        ;       } else {
        ;           qb->next = line;
        ;       }
        ;       ifBLIN_QX4("qbuf new line %u", line->num);
        ;       goto fqba;
        ;   }
        ;   free(file);
        ;   free(func);
        ;   free(line);
            fqba:;
        }
        if  (mife_ctlsize(cfg->tfile) == twind->len) break;
    }
    if  (cfg->flags & BLINREAD_P_D) {
        struct tm *h = NULL;
        time_t     t;

        t = qbuf.time1 / BLINr_E9;
        h = localtime((const time_t*)&t);
        if  (qbuf.time1 == qbuf.time2) {
            printf( "datime %u-%02u-%02u %02u:%02u:%02u.%09u\n"
                  , h->tm_year + 1900
                  , h->tm_mon + 1
                  , h->tm_mday
                  , h->tm_hour
                  , h->tm_min
                  , h->tm_sec
                  , (u_int32_t)(qbuf.time1 % BLINr_E9)
                  );
        } else {
            printf( "datime from %u-%02u-%02u %02u:%02u:%02u.%09u"
                  , h->tm_year + 1900
                  , h->tm_mon + 1
                  , h->tm_mday
                  , h->tm_hour
                  , h->tm_min
                  , h->tm_sec
                  , (u_int32_t)(qbuf.time1 % BLINr_E9)
                  );
            t = qbuf.time2 / BLINr_E9;
            h = localtime((const time_t*)&t);
            printf( " to %u-%02u-%02u %02u:%02u:%02u.%09u\n"
                  , h->tm_year + 1900
                  , h->tm_mon + 1
                  , h->tm_mday
                  , h->tm_hour
                  , h->tm_min
                  , h->tm_sec
                  , (u_int32_t)(qbuf.time1 % BLINr_E9)
                  );
    }   }
    if  (cfg->flags & BLINREAD_P_T) {
        if  (qbuf.time1 == qbuf.time2) {
            printf( "time %"BLIN_O"u.%09u\n"
                  , (qbuf.time1 - fstime) / BLINr_E9
                  , (u_int32_t)((qbuf.time1 - fstime) % BLINr_E9)
                  );
        } else {
            printf( "time from %"BLIN_O"u.%09u to %"BLIN_O"u.%09u\n"
                  , (qbuf.time1 - fstime) / BLINr_E9
                  , (u_int32_t)((qbuf.time1 - fstime) % BLINr_E9)
                  , (qbuf.time2 - fstime) / BLINr_E9
                  , (u_int32_t)((qbuf.time2 - fstime) % BLINr_E9)
                  );
    }   }
    if  (cfg->flags & BLINREAD_P_I) {
        printf("\npid:\n");
        for (BLIN_list_u4 *i = qbuf.pid ; i; i = i->next) printf("%u ", i->num);
        printf("\n");
    }
    if  (cfg->flags & BLINREAD_P_L) {
        printf("\nlevel %02X\n", qbuf.level);
    }
    if  (cfg->flags & BLINREAD_P_R) {
        printf("\nprog:\n");
        for (BLIN_list_pc  *i = qbuf.prog; i; i = i->next) printf("%s ", i->name);
        printf("\n");
    }
    if  (cfg->flags & (BLINREAD_P_S | BLINREAD_P_U | BLINREAD_P_N)) {
        printf("\n");
        p = 0;
        for (BLIN_qbuf_list *file = qbuf.file; file; file = file->next) {
            p = 0;
            for (BLIN_qbuf_list *func = file->list; func; func = func->next) {
                p &= ~(u_int32_t)2;
                for (BLIN_list_u4 *line = func->list; line; line = line->next) {
                    if  (cfg->flags & BLINREAD_P_S) {
                        if  (!(cfg->flags & (BLINREAD_P_U | BLINREAD_P_N))) {
                            printf("%s\n", file->name);
                            break;
                        } else if (p & 1) {
                            printf("%.*s ", (int)strlen(file->name), BLINr_SPACES);
                        } else {
                            printf("%s:", file->name);
                            p |= 1;
                    }   }
                    if  (cfg->flags & BLINREAD_P_U) {
                        if  (!(cfg->flags & BLINREAD_P_N)) {
                            printf("%s\n", func->name);
                            break;
                        } else if  (p & 2) {
                            printf("%.*s ", (int)strlen(func->name), BLINr_SPACES);
                        } else {
                            printf("%s:", func->name);
                            p |= 2;
                    }   }
                    if  (cfg->flags & BLINREAD_P_N) printf("%u\n", line->num);
    }   }   }   }
out:
    return(ex);
#   undef isJX
#   undef isJW
#   undef blin_internal_flags
}
